require 'securerandom'
require 'yaml'

module ProcessMetrics
  class Base
    attr_accessor :data, :uuid, :parent_uuid

    def initialize(name, parent=nil)
      @uuid        = SecureRandom.uuid
      @parent_uuid = parent ? parent.uuid : nil
      @data        = nil
      @name        = name
      @started_at  = Time.now
      @finished_at = nil
    end

    def attributes
      {
        uuid: @uuid,
        parent_uuid: @parent_uuid,
        name: @name,
        data: YAML::dump(@data),
        started_at: @started_at ? @started_at.strftime("%Y-%m-%d %H:%M:%S.%N") : nil,
        finished_at: @finished_at ? @finished_at.strftime("%Y-%m-%d %H:%M:%S.%N") : nil
      }
    end

    def finish
      @finished_at = Time.now
    end

    def finish
      @finished_at = Time.now
    end

    def save
      ProcessMetrics.config.persistence.save(self)
    end

    def measure(name, &block)
      work(name, self, &block)
    end

    def self.work(name, parent=nil, &block)
      metric = ProcessMetrics::Base.new name, parent
      result = nil
      begin
        result = block.call metric
        metric.finish
      ensure
        Thread.new(metric) { |_metric| _metric.save }.join
      end
      result
    end

    def to_s
      self.class.name + " " + attributes.to_s
    end
  end
end
