module ProcessMetrics
  module ClassMethods
    @@replacing = false

    def measure(*methods_names)
      methods_names.flatten!
      options = extract_options(methods_names)

      methods_names.each do |method_name|
        define_wrapper_method(method_name, options)
        wrap_method(method_name) if method_exists?(method_name)
      end
    end

    def method_added(method_name)
      return if method_name =~ /^measure_/ || method_name =~ /^raw_/
      ProcessMetrics.config.logger.debug "Adding method #{method_name} in #{self}"

      measure_method_name = :"measure_#{method_name}"

      if self.instance_methods.include? measure_method_name
        ProcessMetrics.config.logger.debug "#{self}##{measure_method_name} exists. Wrapping..."
        wrap_method(method_name)
      else
        ProcessMetrics.config.logger.debug "#{self}##{measure_method_name} does not exist. Exiting..."
      end
    end

    private
    def extract_options(array)
      array.last && array.last.is_a?(Hash) ? array.pop : nil
    end

    def define_wrapper_method(method_name, options)
      measure_method_name = :"measure_#{method_name}"
      raw_method_name     = :"raw_#{method_name}"

      self.send(:define_method, measure_method_name) do |*args, &block|
        parent       = nil
        uuid         = extract_option(options, :uuid)
        parent_uuid  = extract_option(options, :parent_uuid)
        process_name = extract_option(options, :name) || "#{self.class.name}##{method_name}"

        if parent_uuid
          parent       = ProcessMetrics::Base.new(process_name)
          parent.uuid  = parent_uuid
        end

        ProcessMetrics.measure(process_name, parent) do |metric|
          ProcessMetrics.config.logger.debug "About to send #{raw_method_name} with args #{args.inspect} and block #{block.inspect} to #{self}"
          metric.uuid = uuid if uuid
          metric.data = {args: args, block: block}

          send(raw_method_name, *args, &block)
        end
      end
    end

    def method_exists?(method_name)
      self.instance_methods.include? method_name
    end

    def wrap_method(method_name)
      return if @@replacing

      measure_method_name = :"measure_#{method_name}"
      raw_method_name     = :"raw_#{method_name}"

      @@replacing = true
      alias_method raw_method_name, method_name
      alias_method method_name, measure_method_name
      @@replacing = false
      nil
    end
  end

  module Timer
    def self.included(base)
      base.extend ProcessMetrics::ClassMethods
    end

    def extract_option(options, option_name)
      return unless options && options[option_name]

      if options[option_name].respond_to?(:call)
        # It's a proc. parent_uuid is the return value
        options[option_name].call(self)
      elsif options[option_name].is_a?(Symbol)
        # It's a symbol. Call method on object and parent_uuid is its result.
        send(options[option_name])
      else
        # It's something static. Just call #to_s
        options[option_name].to_s
      end
    end
  end
end
